from datetime import datetime

import requests

from desec_dns_api import UnexpectedError, FieldError, UnauthorizedError, NonEmptyAccountError
from desec_dns_api.tokens import Token


class Captcha:
    def __init__(self, id: str, challenge: str):
        self.id = id
        self.challenge = challenge

    def __str__(self):
        return self.challenge


class Account:
    def __init__(self, id: str, created: datetime, email: str, limit_domains: int):
        self.id = id
        self.created = created
        self.email = email
        self.limit_domains = limit_domains

    def __str__(self):
        return self.email


def obtain_captcha(base_url: str = 'https://desec.io/api') -> Captcha:
    """
    Gets a captcha from desec API.
    https://desec.readthedocs.io/en/latest/auth/account.html#obtain-a-captcha

    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: The Captcha with a challenge to solve.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.post(f'{ base_url }/v1/captcha/')

    if r.status_code != 201:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    # {
    #   "id": "f743336e-uuid-uuid-uuid-70cdfd16a930",
    #   "challenge": "iVBORw0KGgoAAAANSUhEUgAAAKAAAAA8C....fFUvrwAAAABJRU5ErkJggg=="
    # }
    return Captcha(
        id=r.json()['id'],
        challenge=r.json()['challenge']
    )


def register(email: str, password: str, captcha: Captcha, solution: str, domain: str = None, base_url: str = 'https://desec.io/api'):
    """
    Registers an account with desec API.
    https://desec.readthedocs.io/en/latest/auth/account.html#register-account

    :param email: Email address for the account
    :param password: Password for the account
    :param captcha: Captcha object
    :param solution: Solution to the captcha
    :param domain: DNS zone that should be created with the account
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :raises FieldError: If HTTP 400 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {
        'email': email,
        'password': password,
        'captcha': {
            'id': captcha.id,
            'solution': solution
        }
    }
    if domain:
        data['domain'] = domain

    r = requests.post(f'{ base_url }/v1/auth/', json=data)

    if r.status_code == 400:
        msg = ''
        for field in r.json().keys():
            msg += f'Errors in field "{ field }": { " ".join([e for e in r.json()[field]]) }\n'
        raise FieldError(msg.strip())
    elif r.status_code != 202:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')


def login(email: str, password: str, base_url: str = 'https://desec.io/api') -> Token:
    """
    Login and retrieve auth token.
    https://desec.readthedocs.io/en/latest/auth/account.html#log-in

    :param email: Account email address
    :param password: Account password
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: Authentication token
    :raises UnauthorizedError: If HTTP 401 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {
        'email': email,
        'password': password
    }

    r = requests.post(f'{ base_url }/v1/auth/login/', json=data)

    if r.status_code == 401:
        raise UnauthorizedError()
    elif r.status_code != 200:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    # {
    #   "id": "868e3601-uuid-uuid-uuid-ebad1abb4476",
    #   "created": "1970-01-01T23:06:52.847327Z",
    #   "last_used": "1970-01-01T23:06:52.847327Z",  # could be None
    #   "name": "login",
    #   "token": "My_superSecretToken"
    # }
    return Token(
        id=r.json()['id'],
        created=datetime.strptime(r.json()['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        last_used=datetime.strptime(r.json()['last_used'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z') if r.json()['last_used'] else None,
        name=r.json()['name'],
        token=r.json()['token']
    )


def info(token: str, base_url: str = 'https://desec.io/api') -> Account:
    """
    Retrieve account info.
    https://desec.readthedocs.io/en/latest/auth/account.html#retrieve-account-information

    :param token: Authentication token
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: Account info
    :raises UnauthorizedError: If HTTP 401 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.get(f'{ base_url }/v1/auth/account/', headers={'Authorization': f'Token { token }'})

    if r.status_code == 401:
        raise UnauthorizedError()
    elif r.status_code != 200:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    # {
    #   "created": "2019-10-16T18:09:17.715702Z",
    #   "email": "youremailaddress@example.com",
    #   "id": "9ab16e5c-805d-4ab1-9030-af3f5a541d47",
    #   "limit_domains": 5
    # }
    return Account(
        id=r.json()['id'],
        created=datetime.strptime(r.json()['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        email=r.json()['email'],
        limit_domains=r.json()['limit_domains']
    )


def reset_password(email: str, captcha: Captcha, solution: str, base_url: str = 'https://desec.io/api'):
    """
    Request a new password.
    https://desec.readthedocs.io/en/latest/auth/account.html#password-reset

    :param email: Email address of the account
    :param captcha: Captcha object
    :param solution: Solution to the solved captcha
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :raises FieldError: If HTTP 400 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {
        'email': email,
        'captcha': {
            'id': captcha.id,
            'solution': solution
        }
    }

    r = requests.post(f'{ base_url }/v1/auth/account/reset-password/', json=data)

    if r.status_code == 400:
        msg = ''
        for field in r.json()['captcha']:
            msg += f'Errors in field "{ field }": { " ".join([e for e in r.json()["captcha"][field]]) }\n'
        raise FieldError(msg.strip())
    elif r.status_code != 202:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')


def change_email_address(email: str, password: str, new_email: str, base_url: str = 'https://desec.io/api'):
    """
    Request email address change.
    https://desec.readthedocs.io/en/latest/auth/account.html#change-email-address

    :param email: Email address of the account
    :param password: Password of the account
    :param new_email: New email address for the account
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {
        'email': email,
        'password': password,
        'new_email': new_email
    }

    r = requests.post(f'{ base_url }/v1/auth/account/change-email/', json=data)

    if r.status_code != 202:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')


def delete_account(email: str, password: str, base_url: str = 'https://desec.io/api'):
    """
    Request account deletion.
    https://desec.readthedocs.io/en/latest/auth/account.html#delete-account

    :param email: Email address of the account
    :param password: Password of the account
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :raises NonEmptyAccountError: If HTTP 409 is returned.
    :raises UnauthorizedError: If HTTP 403 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {
        'email': email,
        'password': password
    }

    r = requests.post(f'{ base_url }/v1/auth/account/delete/', json=data)

    if r.status_code == 409:
        raise NonEmptyAccountError()
    elif r.status_code == 403:
        try:
            raise UnauthorizedError(r.json()['detail'])
        except (ValueError, KeyError):
            # no JSON returned
            raise UnauthorizedError()
    elif r.status_code != 202:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')


def logout(token: str, base_url: str = 'https://desec.io/api'):
    """
    Logout and remove token from account.
    https://desec.readthedocs.io/en/latest/auth/account.html#log-out

    :param token: The token to remove.
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.post(f'{ base_url }/v1/auth/logout/', headers={'Authorization': f'Token { token }'})

    if r.status_code != 204:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

