from typing import List
from datetime import datetime

import requests

from desec_dns_api import UnexpectedError, ClientError


class Key:
    def __init__(self, dnskey: str, flags: int, keytype: str, ds: List[str]):
        self.dnskey = dnskey
        self.flags = flags
        self.keytype = keytype
        self.ds = ds


class Domain:
    def __init__(self, name: str, created: datetime, published: datetime, touched: datetime, minimum_ttl: int, keys: List[Key]):
        self.name = name
        self.created = created
        self.published = published
        self.touched = touched
        self.minimum_ttl = minimum_ttl
        self.keys = keys

    def __str__(self):
        return self.name


def create_domain(token: str, name: str, base_url: str = 'https://desec.io/api') -> Domain:
    """
    Create a new domain.
    https://desec.readthedocs.io/en/latest/dns/domains.html#creating-a-domain

    :param token: The token to authenticate
    :param name: The domain to create
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: Domain that got created
    :raises ClientError: If HTTP 400 or 403 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {
        'name': name
    }
    r = requests.post(f'{ base_url }/v1/domains/', headers={'Authorization': f'Token { token }'}, json=data)

    if r.status_code in [400, 403]:
        # 400: Improper request
        # 403: Domain limit exceeded
        try:
            raise ClientError(r.json()['detail'])
        except (ValueError, KeyError):
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')
    elif r.status_code != 201:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    # {
    #     "created": "2018-09-18T16:36:16.510368Z",
    #     "keys": [
    #         {
    #             "dnskey": "257 3 13 WFRl60...",
    #             "ds": [
    #                 "6006 13 1 8581e9...",
    #                 "6006 13 2 f34b75...",
    #                 "6006 13 3 dfb325...",
    #                 "6006 13 4 2fdcf8..."
    #             ],
    #             "flags": 257,
    #             "keytype": "csk"
    #         },
    #         ...
    #     ],
    #     "minimum_ttl": 3600,
    #     "name": "example.com",
    #     "published": "2018-09-18T17:21:38.348112Z",
    #     "touched": "2018-09-18T17:21:38.348112Z"
    # }
    return Domain(
        name=r.json()['name'],
        created=datetime.strptime(r.json()['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        published=datetime.strptime(r.json()['published'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        touched=datetime.strptime(r.json()['touched'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        minimum_ttl=r.json()['minimum_ttl'],
        keys=[Key(
            dnskey=k['dnskey'],
            flags=k['flags'],
            keytype=k['keytype'],
            ds=k['ds']
        ) for k in r.json()['keys']]
    )


def list_domains(token: str, base_url: str = 'https://desec.io/api') -> List[Domain]:
    """
    List all domains.
    https://desec.readthedocs.io/en/latest/dns/domains.html#listing-domains

    :param token: The token to authenticate
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: Domains in your account
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    # TODO: handle pagination: https://desec.readthedocs.io/en/latest/dns/rrsets.html#pagination
    r = requests.get(f'{ base_url }/v1/domains/?cursor=', headers={'Authorization': f'Token { token }'})

    if r.status_code != 200:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    return [Domain(
        name=d['name'],
        created=datetime.strptime(d['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        published=datetime.strptime(d['published'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        touched=datetime.strptime(d['touched'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        minimum_ttl=d['minimum_ttl'],
        keys=[]
    ) for d in r.json()]


def get_domain(token: str, name: str, base_url: str = 'https://desec.io/api') -> Domain:
    """
    Retrieve a domain with a specific name.
    https://desec.readthedocs.io/en/latest/dns/domains.html#retrieving-a-specific-domain

    :param token: The token to authenticate
    :param name: Name of the domain
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: Domain object
    :raises ClientError: If HTTP 404 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.get(f'{ base_url }/v1/domains/{ name }/', headers={'Authorization': f'Token { token }'})

    if r.status_code == 404:
        try:
            raise ClientError(r.json()['detail'])
        except (ValueError, KeyError):
            raise UnexpectedError(f'API returned HTTP{r.status_code}')
    if r.status_code != 200:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    return Domain(
        name=r.json()['name'],
        created=datetime.strptime(r.json()['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        published=datetime.strptime(r.json()['published'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        touched=datetime.strptime(r.json()['touched'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        minimum_ttl=r.json()['minimum_ttl'],
        keys=[Key(
            dnskey=k['dnskey'],
            flags=k['flags'],
            keytype=k['keytype'],
            ds=k['ds']
        ) for k in r.json()['keys']]
    )


def delete_domain(token: str, name: str, base_url: str = 'https://desec.io/api'):
    """
    Delete a domain.
    https://desec.readthedocs.io/en/latest/dns/domains.html#deleting-a-domain

    :param token: The token to authenticate
    :param name: The domain to delete
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.delete(f'{ base_url }/v1/domains/{ name }/', headers={'Authorization': f'Token { token }'})

    if r.status_code != 204:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

