import re
from typing import List
from datetime import datetime

import requests

from desec_dns_api import UnexpectedError, ClientError


class RRset:
    def __init__(self, subname: str, domain: str, created: datetime, type: str, ttl: int, records: List[str]):
        self.name = f'{ subname }.{ domain }.'
        self.subname = subname
        self.domain = domain
        self.created = created
        self.type = type
        self.ttl = ttl
        self.records = records

    def __str__(self):
        return '\n'.join([f'{ self.name } { self.ttl } IN { self.type } { r }' for r in self.records])


def create_rrset(token: str, name: str, subname: str, type: str, ttl: int, records: List[str], base_url: str = 'https://desec.io/api') -> RRset:
    """
    Create a new recordset for a domain.
    https://desec.readthedocs.io/en/latest/dns/rrsets.html#creating-an-rrset

    :param token: The token to authenticate
    :param name: Name of the domain to use
    :param subname: Subdomain to use
    :param type: Record type
    :param ttl: Record TTL
    :param records: List of records
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: Record that got created
    :raises ClientError: If HTTP 400 or 429 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {
        'subname': subname,
        'type': type,
        'ttl': ttl,
        'records': records
    }
    # TXT records must be enclosed in double quotes
    # https://desec.readthedocs.io/en/latest/dns/rrsets.html#caveats
    if type == 'TXT':
        data['records'] = [f'"{ r }"' for r in records]
    r = requests.post(f'{ base_url }/v1/domains/{ name }/rrsets/',
                      headers={'Authorization': f'Token { token }'}, json=data)

    if r.status_code in [400, 429]:
        # 400: Operation cannot be performed with the given parameters.
        # 429: Request was throttled. Expected available in 1 second.
        raise ClientError(r.json()['detail'])
    elif r.status_code != 201:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    # {
    #     "created": "2019-09-18T16:32:16.510368Z",
    #     "domain": "example.com",
    #     "subname": "www",
    #     "name": "www.example.com.",
    #     "type": "A",
    #     "records": [
    #         "127.0.0.1",
    #         "127.0.0.2"
    #     ],
    #     "ttl": 3600
    # }
    return RRset(
        subname=r.json()['subname'],
        domain=r.json()['domain'],
        created=datetime.strptime(r.json()['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        type=r.json()['type'],
        ttl=r.json()['ttl'],
        records=r.json()['records']
    )


# TODO: Bulk Creation of RRsets https://gitlab.com/XenGi/desec-dns-cli/issues/2
# https://desec.readthedocs.io/en/latest/dns/rrsets.html#bulk-creation-of-rrsets


def get_rrsets(token: str, name: str, _query_param='', base_url: str = 'https://desec.io/api') -> List[RRset]:
    """
    Get all record sets in a zone.
    https://desec.readthedocs.io/en/latest/dns/rrsets.html#retrieving-all-rrsets-in-a-zone

    :param token: The token to authenticate
    :param name: Name of the domain to use
    :param _query_param: Additional parameter for URL
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: List of found Records
    :raises ClientError: If HTTP 404 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    if _query_param != '':
        _query_param = f'{ _query_param }&'
    r = requests.get(f'{ base_url }/v1/domains/{ name }/rrsets/?{ _query_param }cursor=',
                     headers={'Authorization': f'Token { token }'})

    if r.status_code != 200:
        try:
            if r.status_code == 404:
                # 404: Not found
                raise ClientError(r.json()['detail'])
            else:
                raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    # https://desec.readthedocs.io/en/latest/dns/rrsets.html#pagination
    rrsets = []
    while 'Link' in r.headers.keys():
        for rrset in r.json():
            rrsets.append(RRset(
                subname=rrset['subname'],
                domain=rrset['domain'],
                created=datetime.strptime(rrset['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
                type=rrset['type'],
                ttl=rrset['ttl'],
                records=rrset['records']
            ))
        # Link: <https://desec.io/api/v1/domains/{domain}/rrsets/?cursor=>; rel="first",
        #   <https://desec.io/api/v1/domains/{domain}/rrsets/?cursor=:prev_cursor>; rel="prev",
        #   <https://desec.io/api/v1/domains/{domain}/rrsets/?cursor=:next_cursor>; rel="next"
        m = re.search(r'<.+>; rel="first"(,\s+<(?P<previous>.+)>; rel="prev")?(,\s+<(?P<next>.+)>; rel="next")?',
                      r.headers['Link'])
        if not m.group('next'):
            # reached the last page
            return rrsets

        r = requests.get(m.group('next'), headers={'Authorization': f'Token { token }'})
        if r.status_code != 200:
            try:
                raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
            except (ValueError, KeyError):
                # no JSON returned
                raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    return [RRset(
        subname=rs['subname'],
        domain=rs['domain'],
        created=datetime.strptime(rs['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        type=rs['type'],
        ttl=rs['ttl'],
        records=rs['records']
    ) for rs in r.json()]


def get_rrsets_by_type(token: str, name: str, type: str, base_url: str = 'https://desec.io/api') -> List[RRset]:
    """
    Retrieve an array of all RRsets from your zone that have a specific type.
    https://desec.readthedocs.io/en/latest/dns/rrsets.html#filtering-by-record-type

    :param token: The token to authenticate
    :param name: Name of the domain to use
    :param type: Type to filter for
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: List of found Records
    :raises ClientError: If HTTP 404 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    return get_rrsets(token=token, name=name, _query_param=f'type={ type }', base_url=base_url)


def get_rrsets_by_subname(token: str, name: str, subname: str, base_url: str = 'https://desec.io/api') -> List[RRset]:
    """
    Retrieve an array of all RRsets from your zone that have a specific subname.
    https://desec.readthedocs.io/en/latest/dns/rrsets.html#filtering-by-subname

    :param token: The token to authenticate
    :param name: Name of the domain to use
    :param subname: Subname to filter for
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: List of found Records
    :raises ClientError: If HTTP 404 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    return get_rrsets(token=token, name=name, _query_param=f'subname={ subname }', base_url=base_url)


def get_rrset(token: str, name: str, subname: str, type: str, base_url: str = 'https://desec.io/api') -> RRset:
    """
    Retrieve a RRset with a specific name and type from your zone.
    https://desec.readthedocs.io/en/latest/dns/rrsets.html#retrieving-a-specific-rrset

    :param token: The token to authenticate
    :param name: Name of the domain to use
    :param subname: Subname to filter for
    :param type: Type to filter for
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: List of found Records
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.get(f'{ base_url }/v1/domains/{ name }/rrsets/{ subname }/{ type }/',
                     headers={'Authorization': f'Token { token }'})

    if r.status_code != 200:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    return RRset(
        subname=r.json()['subname'],
        domain=r.json()['domain'],
        created=datetime.strptime(r.json()['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        type=r.json()['type'],
        ttl=r.json()['ttl'],
        records=r.json()['records']
    )


def modify_rrset(token: str, name: str, subname: str, type: str, ttl: int = None, records: List[str] = None, base_url: str = 'https://desec.io/api') -> RRset:
    """
    Change the ttl or records of a Record Set.
    https://desec.readthedocs.io/en/latest/dns/rrsets.html#modifying-an-rrset

    :param token: The token to authenticate
    :param name: Name of the domain to use
    :param subname: Subname to filter for
    :param type: Type to filter for
    :param ttl: TTL of the RRset
    :param records: List of records in the RRset
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: Modified Record
    :raises ClientError: If HTTP 400 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {}
    if ttl:
        data['ttl'] = ttl
    if records:
        data['records'] = records
        # TXT records must be enclosed in double quotes
        # https://desec.readthedocs.io/en/latest/dns/rrsets.html#caveats
        if type == 'TXT':
            data['records'] = [f'"{ r }"' for r in records]
    r = requests.patch(f'{ base_url }/v1/domains/{ name }/rrsets/{ subname }/{ type }/',
                       headers={'Authorization': f'Token { token }'}, json=data)

    if r.status_code == 400:
        try:
            raise ClientError(r.json())
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')
    elif r.status_code != 200:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    return RRset(
        subname=r.json()['subname'],
        domain=r.json()['domain'],
        created=datetime.strptime(r.json()['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        type=r.json()['type'],
        ttl=r.json()['ttl'],
        records=r.json()['records']
    )


# TODO: Bulk Modification of RRsets https://gitlab.com/XenGi/desec-dns-cli/issues/4
# https://desec.readthedocs.io/en/latest/dns/rrsets.html#bulk-modification-of-rrsets


def delete_rrset(token: str, name: str, subname: str, type: str, base_url: str = 'https://desec.io/api'):
    """
    Delete a RRset.
    https://desec.readthedocs.io/en/latest/dns/rrsets.html#deleting-an-rrset

    :param token: The token to authenticate
    :param name: Name of the domain to use
    :param subname: Subname to filter for
    :param type: Type to filter for
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.delete(f'{ base_url }/v1/domains/{ name }/rrsets/{ subname }/{ type }/',
                        headers={'Authorization': f'Token { token }'})

    if r.status_code != 204:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')


# TODO: Bulk Deletion of RRsets https://gitlab.com/XenGi/desec-dns-cli/issues/5
# https://desec.readthedocs.io/en/latest/dns/rrsets.html#bulk-deletion-of-rrsets

