from typing import List
from datetime import datetime

import requests

from desec_dns_api import UnauthorizedError, UnexpectedError


class Token:
    def __init__(self, id: str, created: datetime, last_used: datetime, name: str, token: str):
        self.id = id
        self.created = created
        self.last_used = last_used
        self.name = name
        self.token = token

    def __str__(self):
        return self.token


def get_all_tokens(token: str, base_url: str = 'https://desec.io/api') -> List[Token]:
    """
    Retrieves all current tokens.
    https://desec.readthedocs.io/en/latest/auth/tokens.html#retrieving-all-current-tokens

    :param token: The token to authenticate
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: List of tokens
    :raises UnauthorizedError: If HTTP 401 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.get(f'{ base_url }/v1/auth/tokens/', headers={'Authorization': f'Token { token }'})

    if r.status_code == 401:
        raise UnauthorizedError()
    elif r.status_code != 200:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    # [
    #     {
    #         "created": "2018-09-06T07:05:54.080564Z",
    #         "id": "3159e485-5499-46c0-ae2b-aeb84d627a8e",
    #         "name": "login"
    #     },
    #     {
    #         "created": "2018-09-06T08:53:26.428396Z",
    #         "id": "76d6e39d-65bc-4ab2-a1b7-6e94eee0a534",
    #         "name": ""
    #     }
    # ]
    return [Token(
        id=t['id'],
        created=datetime.strptime(t['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        last_used=datetime.strptime(t['last_used'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z') if t['last_used'] else None,
        name=t['name'],
        token=''
    ) for t in r.json()]


def create_token(token: str, name: str = '', base_url: str = 'https://desec.io/api') -> Token:
    """
    Create a new token for the account.
    https://desec.readthedocs.io/en/latest/auth/tokens.html#create-additional-tokens

    :param token: The token to authenticate
    :param name: Optional name for the token
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :return: The created token
    :raises UnauthorizedError: If HTTP 401 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    data = {
        'name': name
    }
    r = requests.post(f'{ base_url }/v1/auth/tokens/', headers={'Authorization': f'Token { token }'}, json=data)

    if r.status_code == 401:
        raise UnauthorizedError()
    elif r.status_code != 201:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

    # {
    #     "created": "2018-09-06T09:08:43.762697Z",
    #     "id": "3a6b94b5-d20e-40bd-a7cc-521f5c79fab3",
    #     "token": "4pnk7u+NHvrEkFzrhFDRTjGFyX+S",
    #     "name": "my new token"
    # }
    return Token(
        id=r.json()['id'],
        created=datetime.strptime(r.json()['created'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z'),
        last_used=datetime.strptime(r.json()['last_used'].replace('Z', 'UTC'), '%Y-%m-%dT%H:%M:%S.%f%Z') if r.json()['last_used'] else None,
        name=r.json()['name'],
        token=r.json()['token']
    )


def delete_token(token: str, id: str, base_url: str = 'https://desec.io/api'):
    """
    Remove a token from the account.
    https://desec.readthedocs.io/en/latest/auth/tokens.html#delete-tokens

    :param token: The token to authenticate
    :param id: The token id to delete
    :param base_url: URL to desec API endpoint (default: https://desec.io/api)
    :raises UnauthorizedError: If HTTP 401 is returned.
    :raises UnexpectedError: If JSON data returns no detail key or invalid JSON.
    """
    r = requests.delete(f'{ base_url }/v1/auth/tokens/{ id }/', headers={'Authorization': f'Token { token }'})

    if r.status_code == 401:
        raise UnauthorizedError()
    elif r.status_code != 204:
        try:
            raise UnexpectedError(f'API returned HTTP{ r.status_code }: { r.json()["detail"] }')
        except (ValueError, KeyError):
            # no JSON returned
            raise UnexpectedError(f'API returned HTTP{ r.status_code }')

